/*
 * Basic utility procedures for dealing with cubes and arrays
 */
#include <stdio.h>
#include "logic.h"

char *alloc();

/*
 * Allocate space for a cube with `nv' terms
 */
struct cube *
newcube(nv)
int nv;
{
	struct cube *p;

	p = (struct cube *)alloc(sizeof(struct cube) + (nv-1)*sizeof(char));
	return p;
}

/*
 * Set cube's elements f .. f+n-1 to v.
 */
setcube(p, f, n, v)
register struct cube *p;
int f, n, v;
{
	register int i;
	register char *lp;

	lp = p->c_logic + f;
	for( i = n; i > 0; --i )
		*lp++ = v;
}

struct arraylist hfrelist;

/*
 * Allocate space for a new array header with `ni' inputs and `nv-ni' outputs
 * The array is initialized to empty.
 */
struct array *
newarray(ni, nv)
int ni, nv;
{
	struct array *p;

	if( (p = hfrelist.l_first) != NULL )
		hfrelist.l_first = p->a_next;
	else
		p = (struct array *)alloc(sizeof(struct array));
	p->a_first = NULL;
	p->a_noin = ni;
	p->a_nvar = nv;
	return(p);
}

/*
 * Add a cube to an array.
 */
add(a, c)
struct array *a;
struct cube *c;
{
	if( a->a_first == NULL )
		a->a_first = c;
	else
		a->a_last->c_next = c;
	a->a_last = c;
	c->c_next = NULL;
}

/*
 * Add an array to an array list.
 */
aadd(l, a)
struct arraylist *l;
struct array *a;
{
	if( l->l_first == NULL )
		l->l_first = a;
	else
		l->l_last->a_next = a;
	l->l_last = a;
	a->a_next = NULL;
}

/*
 * Free all the storage used by an array.
 */
release(a)
struct array *a;
{
	register struct cube *p, *q;

	for( p = a->a_first; p != NULL; p = q ){
		q = p->c_next;
		cfree(p);
	}
	afree(a);
}

char bitnames[4] = { '?', '0', '1', 'X' };

/*
 * Print out an array.
 */
prary(a, f)
struct array *a;
FILE *f;
{
	struct cube *p;
	int i;
	char z;

	if( debug )
		fprintf(f, "first:%x last:%x\n", a->a_first, a->a_last);
	for( p = a->a_first; p != NULL; p = p->c_next ){
		if( debug )
			fprintf(f, "%x: ", p);
		for( i = 0; i < a->a_nvar; ++i ){
			if( i == a->a_noin )
				fprintf(f, " -> ");
			else if (i % 8 == 0)
				fputc(' ', f);
			z = p->c_logic[i];
			fputc(bitnames[z], f);
		}
		fputc('\n', f);
	}
}

/*
 * Header assignment - copy info in header b to header a,
 * release header b & cubes in a.
 */
hass(a, b)
struct array *a, *b;
{
	struct cube *p, *q;

	for( p = a->a_first; p != NULL; p = q ){
		q = p->c_next;
		cfree(p);
	}
	a->a_first = b->a_first;
	a->a_last = b->a_last;
	a->a_noin = b->a_noin;
	a->a_nvar = b->a_nvar;
	afree(b);
}

/*
 * Append array b to array a; free header b.
 */
append(a, b)
struct array *a, *b;
{
	arr_compat(a, b);
	if( b->a_first != NULL ){
		if( a->a_first == NULL )
			a->a_first = b->a_first;
		else
			a->a_last->c_next = b->a_first;
		a->a_last = b->a_last;
	}
	afree(b);
}

/* Copy the logic part of a cube */
ccopy(p, q, nv)
struct cube *p, *q;
int nv;
{
	int i;

	for( i = 0; i < nv; ++i )
		p->c_logic[i] = q->c_logic[i];
}

/*
 * Make a copy of an array.
 */
struct array *
acopy(a)
struct array *a;
{
	struct cube *p, **q0, *q;
	register char *s1, *s2;
	struct array *b;
	register int i;

	b = newarray(a->a_noin, a->a_nvar);
	q0 = &b->a_first;
	for( p = a->a_first; p != NULL; p = p->c_next ){
		q = newcube(a->a_nvar);
		s1 = p->c_logic;
		s2 = q->c_logic;
		for( i = a->a_nvar; i > 0; --i )
			*s2++ = *s1++;
		*q0 = q;
		q0 = &q->c_next;
	}
	*q0 = NULL;
	b->a_last = q;
	return b;
}

/*
 * Change column i of an array.
 * Z's are changed to z, O's to o, and X's to x.
 */
chcol(a, i, z, o, x)
struct array *a;
int i;
char z, o, x;
{
	struct cube *p;
	register char *s;

	for( p = a->a_first; p != NULL; p = p->c_next ){
		s = &p->c_logic[i];
		*s = *s==Z? z: *s==O? o: x;
	}
}

/* Complain if two arrays are not compatible */
incompat()
{
	fprintf(stderr, "!!Incompatible arrays\n");
	exit(1);
}

